VERSION 5.00
Begin VB.Form frmMain 
   BorderStyle     =   0  'None
   Caption         =   "Form1"
   ClientHeight    =   3195
   ClientLeft      =   0
   ClientTop       =   0
   ClientWidth     =   4680
   LinkTopic       =   "Form1"
   ScaleHeight     =   213
   ScaleMode       =   3  'Pixel
   ScaleWidth      =   312
   ShowInTaskbar   =   0   'False
   StartUpPosition =   3  'Windows Default
   WindowState     =   2  'Maximized
End
Attribute VB_Name = "frmMain"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = False
Attribute VB_PredeclaredId = True
Attribute VB_Exposed = False
'--------------------------------------------------------------------------------------
' REAL-TIME IMAGE MANIPULATION IN VISUAL BASIC
' 2001 by Thomas J. Hatfield (Digital Construct)
' Bump-mapping algorithm by Unlimited Realities
'
' This program demonstrates how to access the raw pixel data of a Picture object
' using Visual Basic pointers.  If anyone has ever told you that VB does not support
' pointers, they were lying.  :)  We're going to prove that high-speed image manipu-
' lation is possible in Visual Basic, and we're going to do it by digging into the
' most basic parts of Windows programming:  the Component Object Model (COM).
'
' A basic understanding of COM will help immensely in understanding this program.
' For more information on how this program works (and where the idea originated),
' read a section called "Dealing With Pointers" in chapter two of Hardcore Visual
' Basic: Second Edition, a book written by Bruce McKinney.  You'll find a copy in
' the Books section of MSDN-1997 or later.
'
' This source code is fully and articulately documented, but if you need help
' interpreting anything, you may contact me at 'hurendo_kun@hotmail.com'.
'
' The bump-mapping algorithm used in this program was taken directly from a similar
' (but poorly-documented) example created by Unlimited Realities.  For more examples
' of what you can do with this technique, visit their website:
'
'       http://www.ur.co.nz/urcorp/default.asp
'--------------------------------------------------------------------------------------

Option Explicit

' The following user-defined types comply with Windows COM safe-array descriptions,
' which are used intrinsically by Visual Basic.  We're building our own safe-array
' so we can access individual elements of the array descriptor, which will be
' necessary when we use pointers.

Private Type SAFEARRAYBOUND
    cElements As Long
    lLbound As Long
End Type

Private Type SAFEARRAY1D
    cDims As Integer
    fFeatures As Integer
    cbElements As Long
    cLocks As Long
    pvData As Long
    Bounds(0 To 0) As SAFEARRAYBOUND
End Type

Private Type SAFEARRAY2D
    cDims As Integer
    fFeatures As Integer
    cbElements As Long
    cLocks As Long
    pvData As Long
    Bounds(0 To 1) As SAFEARRAYBOUND
End Type

' The following user-defined type is the descriptor of a Windows bitmap.  We use this
' to access a bitmap's dimensions, color depth, and raw pixel data.

Private Type BITMAP
        bmType          As Long
        bmWidth         As Long
        bmHeight        As Long
        bmWidthBytes    As Long
        bmPlanes        As Integer
        bmBitsPixel     As Integer
        bmBits          As Long     ' <- This holds a pointer to the pixel data.
End Type

' The next three functions form the basis of this program, as they are used to acquire
' pointers to our images and pass them to other objects.  VarPtrArray is really just
' a copy of VarPtr "overloaded" to accept arrays, which we must do since the image
' data we'll be manipulating will be accessible through an array.
'
' CopyMemory is used to pass a memory pointer between two array descriptors, which
' has the effect of tricking Visual Basic into reading to and writing from a separate
' block of memory using one array in place of another.  We use this technique to
' perform blindingly fast graphics operations.
'
' GetObjectAPI is used to retrieve handles to graphics objects.  In this case, we're
' retrieving a handle to a bitmap so we can access its descriptor (which in turn
' gives us access to its raw pixel data).

Private Declare Function VarPtrArray Lib "msvbvm50.dll" Alias "VarPtr" (Ptr() As Any) As Long
Private Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" (pDst As Any, pSrc As Any, ByVal ByteLen As Long)
Private Declare Function GetObjectAPI Lib "gdi32" Alias "GetObjectA" (ByVal hObject As Long, ByVal nCount As Long, lpObject As Any) As Long

' These two functions are used for performance benchmarking.  We're passing Currency
' instead of LARGE_INTEGER because Currency is the native 64-bit data type in Visual
' Basic, and it's far easier (and safer) than defining our own 64-bit type.

Private Declare Function QueryPerformanceCounter Lib "kernel32" (lpPerformanceCount As Currency) As Long
Private Declare Function QueryPerformanceFrequency Lib "kernel32" (lpFrequency As Currency) As Long

' We need a DirectX object and the usual DirectDraw variables.  We also need a RECT
' to Lock the DirectDraw surface later, so I'm including that here.

Private objDX As New DirectX7
Private objDD As DirectDraw7
Private ddsFront As DirectDrawSurface7
Private ddsBack As DirectDrawSurface7
Private ddpPal As DirectDrawPalette
Private pePal(0 To 255) As PALETTEENTRY
Private rScreen As RECT

' The following object variables will hold the pictures we'll be manipulating.
' We're not loading these onto DirectDraw surfaces because we want to access them
' directly from system memory and dump them onto the back buffer only afterward.
' Technically, you could use DirectDraw surfaces, but this way is faster.  It also
' allows us to load GIF and JPEG images without any modifications to our code!

Private picBump As New StdPicture
Private picGlow As New StdPicture

' The following variables and constants should be self-explanatory.

Private bRunning As Boolean
Private mouseX As Integer, mouseY As Integer
Private curStart As Currency, curEnd As Currency, curFreq As Currency

Private Const scrWidth = 640
Private Const scrHeight = 480
Private Const bitDepth = 8

' These are the images we're using to create the background and phong lightmap.

Private Const fileBump = "texture.gif"  '<- These images must be in your App.Path
Private Const fileGlow = "phong.gif"

Private Sub Form_KeyPress(KeyAscii As Integer)
'   If the user presses a key, end the program:
    bRunning = False
End Sub

Private Sub Form_Load()
'   First, we initialize DirectDraw:
    Call InitDDraw
    
'   Next, we load our images:
    Set picBump = LoadPicture(App.Path & "\" & fileBump)
    Set picGlow = LoadPicture(App.Path & "\" & fileGlow)
    
'   Now we're ready to run the main loop:
    bRunning = True
    Call QueryPerformanceFrequency(curFreq) ' <- This is for the performance test.
    
    Do While bRunning
        Call MainLoop
        DoEvents
    Loop
    
    Call objDD.RestoreDisplayMode
'   DirectDraw clean-up occurs in the Unload function of frmMain.
    Unload Me
End Sub

Private Sub MainLoop()
'   A note before we continue:  all the variables declared locally here can be
'   declared at module-level above if you want a small performance boost.  I'm
'   including them here so you can reference them more easily as you study the main
'   loop.  Feel free to move them to the top of this module.

'   But before anything else, let's start the performance test:
    Call QueryPerformanceCounter(curStart)
    
'   We'll need three byte arrays and two safe-arrays to access pixel data from our
'   Pictures and back buffer.  (We don't need a safe-array for the back buffer.)
    Dim aBump() As Byte, aGlow() As Byte, aDDSBack() As Byte
    Dim saBump As SAFEARRAY2D, saGlow As SAFEARRAY2D

'   We'll also need two bitmap descriptors for our Picture objects and a surface
'   description for the back buffer:
    Dim bmpBump As BITMAP, bmpGlow As BITMAP, ddsd As DDSURFACEDESC2

'--------------------------------------------------------------------------------------
'   Now we're going to create a dynamic real-time bump map.

'   First, we call GetObjectAPI for both Pictures to acquire pointers to the bitmaps
'   they contain.  Basically, this fills our two BITMAP structures with the data in
'   each Picture object.  We'll be using this momentarily.
    Call GetObjectAPI(picBump, Len(bmpBump), bmpBump)
    Call GetObjectAPI(picGlow, Len(bmpGlow), bmpGlow)
    
'   Now we lock our back buffer so we can draw directly to it:
    Call ddsBack.GetSurfaceDesc(ddsd)
    Call ddsBack.Lock(rScreen, ddsd, DDLOCK_WAIT, ByVal 0&)
    
'   Now we configure our three byte arrays to point directly to the pixel data of
'   our back buffer and two Pictures.  To do this for our back buffer, we need to
'   retrieve a pointer using GetLockedArray.

    Call ddsBack.GetLockedArray(aDDSBack)
    
    With saBump
        .cbElements = 1
        .cDims = 2
        .Bounds(0).lLbound = 0
        .Bounds(0).cElements = bmpBump.bmHeight
        .Bounds(1).lLbound = 0
        .Bounds(1).cElements = bmpBump.bmWidthBytes
        .pvData = bmpBump.bmBits
    End With
    Call CopyMemory(ByVal VarPtrArray(aBump), VarPtr(saBump), 4)
    
    With saGlow
        .cbElements = 1
        .cDims = 2
        .Bounds(0).lLbound = 0
        .Bounds(0).cElements = bmpGlow.bmHeight
        .Bounds(1).lLbound = 0
        .Bounds(1).cElements = bmpGlow.bmWidthBytes
        .pvData = bmpGlow.bmBits
    End With
    Call CopyMemory(ByVal VarPtrArray(aGlow), VarPtr(saGlow), 4)
    
    
'   Now we can manipulate the images.  This algorithm was copied directly from a
'   bump-mapping program by Unlimited Realities.  I plan to write my own version of
'   this algorithm after I've had a chance to study it.  In the meantime, this will
'   suffice to demonstrate the technique.

'   Again, these locals can go at the top for a very minor performance boost:
    Dim xdelta As Integer, ydelta As Integer
    Dim lx As Integer, ly As Integer, xtemp As Integer, ytemp As Integer
    Dim c As Integer, r As Integer
    
    lx = -mouseX
    For c = 1 To (scrWidth - 2)
        ly = mouseY
        For r = 1 To (scrHeight - 2)
            xdelta = lx + (CInt(aBump(c - 1, r)) - CInt(aBump(c + 1, r)))
            ydelta = ly + (CInt(aBump(c, r - 1)) - CInt(aBump(c, r + 1)))
            If (ydelta And 65280) Or (xdelta And 65280) Then
                aDDSBack(c, r) = 0
            Else
                aDDSBack(c, r) = aGlow(xdelta, ydelta)
            End If
            ly = ly + 1
        Next
        lx = lx + 1
    Next
    
'   Now we use CopyMemory to clear the safe-array descriptors without erasing the
'   byte arrays (we don't have to do this for the back buffer):
    Call CopyMemory(ByVal VarPtrArray(aBump), 0&, 4)
    Call CopyMemory(ByVal VarPtrArray(aGlow), 0&, 4)
    
'   Unlock the surface:
    Call ddsBack.Unlock(rScreen)

'   Now let's finish the performance test and display the frame-rate:
    Call QueryPerformanceCounter(curEnd)
    Call ddsBack.DrawText(10, 25, Str(1 / ((curEnd - curStart) / curFreq)), False)
    
'   And flip:
    Call ddsFront.Flip(ddsBack, DDFLIP_WAIT)

End Sub

Private Sub InitDDraw()
'   The first thing we need to do is create our DirectDraw object.
'   Before we do that, however, we're going to enumerate display devices and use the
'   last one.  This is because I have two monitors on my system, and I like to run
'   programs on the second while I debug on the first.  If your system has only one
'   monitor, then obviously the last will also be the first, so there's no problem.

'   Create a variable to hold the enumeration, and fill it:
    Dim ddEnum As DirectDrawEnum
    Set ddEnum = objDX.GetDDEnum

'   Create the DirectDraw object:
    Set objDD = objDX.DirectDrawCreate(ddEnum.GetGuid(ddEnum.GetCount))
    'Set objDD = objDX.DirectDrawCreate("")
    Me.Show
    
'   Now we'll set the cooperative level and display mode:
    Call objDD.SetCooperativeLevel(Me.hWnd, DDSCL_FULLSCREEN Or DDSCL_EXCLUSIVE Or DDSCL_ALLOWMODEX)
    Call objDD.SetDisplayMode(scrWidth, scrHeight, bitDepth, 0, DDSDM_DEFAULT)
    
'   Next, we'll create our front and back buffers.  Note that we're including the
'   SYSTEMMEMORY flag in the surface description because we don't want DirectDraw to
'   create the surface in video memory.  Since we'll be writing directly to the
'   surface, using video memory would invoke a SEVERE performance hit.
    
    Dim ddsd As DDSURFACEDESC2
    With ddsd
        .lFlags = DDSD_CAPS Or DDSD_BACKBUFFERCOUNT
        .ddsCaps.lCaps = DDSCAPS_PRIMARYSURFACE Or DDSCAPS_FLIP Or DDSCAPS_COMPLEX Or DDSCAPS_SYSTEMMEMORY
        .ddpfPixelFormat.lFlags = DDPF_PALETTEINDEXED8
        .lBackBufferCount = 1
    End With
    Set ddsFront = objDD.CreateSurface(ddsd)

'   From that, we can attach a back buffer:
    Dim caps As DDSCAPS2
    caps.lCaps = DDSCAPS_BACKBUFFER
    Set ddsBack = ddsFront.GetAttachedSurface(caps)
    
'   We need to create and set a palette for the primary surface.  I'm just using
'   shades of gray here, but you can set the colors to anything you want.
    
    Dim i As Integer
    For i = 0 To 255
        With pePal(i)
            .red = i
            .green = i
            .blue = i
        End With
    Next i
    Set ddpPal = objDD.CreatePalette(DDPCAPS_8BIT Or DDPCAPS_ALLOW256, pePal())
    Call ddsFront.SetPalette(ddpPal)
    
'   Before we continue, let's go ahead and set our screen RECT:
    With rScreen
        .Left = 0
        .Top = 0
        .Right = scrWidth
        .Bottom = scrHeight
    End With
    
'   Finally, set the ForeColor to white:
    Call ddsBack.SetForeColor(vbWhite)
        
End Sub

Private Sub Form_MouseDown(Button As Integer, Shift As Integer, X As Single, Y As Single)
'   If the user clicks the mouse, exit the program:
    bRunning = False
End Sub

Private Sub Form_MouseMove(Button As Integer, Shift As Integer, X As Single, Y As Single)
    mouseX = X - 128
    mouseY = 128 - Y
End Sub

Private Sub Form_Unload(Cancel As Integer)
'   Clean up our DirectDraw objects:
    Set ddsBack = Nothing
    Set ddsFront = Nothing
    Set objDD = Nothing
    
End Sub
